# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
planets
---------

This is a sample dataset build into the engine, this simplifies a few things:

- We can write test scripts using this data, knowing that it will always be available.
- We can write examples using this data, knowing the results will always match.

This data was obtained from:
https://github.com/devstronomy/nasa-data-scraper/blob/master/data/json/planets.json

Licence @ 02-JAN-2022 when copied - MIT Licences attested, but data appears to be
from NASA, which is Public Domain.

To access this dataset you can either run a query against dataset :planets: or you
can instantiate a PlanetData() class and use it like a Relation.

This has a companion dataset, $satellites, to help test joins.
"""
import datetime

from opteryx.constants.attribute_types import OPTERYX_TYPES
from opteryx.models import FlatColumn
from opteryx.models import RelationSchema


def read(end_date=datetime.datetime.utcnow().date()):
    import pyarrow

    # fmt:off
    # Define the data
    data = [
        pyarrow.array([1, 2, 3, 4, 5, 6, 7, 8, 9], type=pyarrow.int64()),
        pyarrow.array(["Mercury", "Venus", "Earth", "Mars", "Jupiter", "Saturn", "Uranus", "Neptune", "Pluto"], type=pyarrow.string()),
        pyarrow.array([0.33, 4.87, 5.97, 0.642, 1898, 568, 86.8, 102, 0.0146], type=pyarrow.float64()),
        pyarrow.array([4879, 12104, 12756, 6792, 142984, 120536, 51118, 49528, 2370], type=pyarrow.int64()),
        pyarrow.array([5427, 5243, 5514, 3933, 1326, 687, 1271, 1638, 2095], type=pyarrow.int64()),
        pyarrow.array([3.7, 8.9, 9.8, 3.7, 23.1, 9, 8.7, 11, 0.7], type=pyarrow.float64()),
        pyarrow.array([4.3, 10.4, 11.2, 5, 59.5, 35.5, 21.3, 23.5, 1.3], type=pyarrow.float64()),
        pyarrow.array([1407.6, -5832.5, 23.9, 24.6, 9.9, 10.7, -17.2, 16.1, -153.3], type=pyarrow.float64()),
        pyarrow.array([4222.6, 2802, 24, 24.7, 9.9, 10.7, 17.2, 16.1, 153.3], type=pyarrow.float64()),
        pyarrow.array([57.9, 108.2, 149.6, 227.9, 778.6, 1433.5, 2872.5, 4495.1, 5906.4], type=pyarrow.float64()),
        pyarrow.array([46, 107.5, 147.1, 206.6, 740.5, 1352.6, 2741.3, 4444.5, 4436.8], type=pyarrow.float64()),
        pyarrow.array([69.8, 108.9, 152.1, 249.2, 816.6, 1514.5, 3003.6, 4545.7, 7375.9], type=pyarrow.float64()),
        pyarrow.array([88, 224.7, 365.2, 687, 4331, 10747, 30589, 59800, 90560], type=pyarrow.float64()),
        pyarrow.array([47.4, 35, 29.8, 24.1, 13.1, 9.7, 6.8, 5.4, 4.7], type=pyarrow.float64()),
        pyarrow.array([7, 3.4, 0, 1.9, 1.3, 2.5, 0.8, 1.8, 17.2], type=pyarrow.float64()),
        pyarrow.array([0.205, 0.007, 0.017, 0.094, 0.049, 0.057, 0.046, 0.011, 0.244], type=pyarrow.float64()),
        pyarrow.array([0.03, 177.4, 23.4, 25.2, 3.1, 26.7, 97.8, 28.3, 122.5], pyarrow.float64()),
        pyarrow.array([167, 464, 15, -63, -108, -139, -197, -201, -225], type=pyarrow.int64()),
        pyarrow.array([0, 92, 1, 0.001, None, None, None, None, 0.00001], pyarrow.float64()),
        pyarrow.array([0, 0, 1, 2, 79, 82, 27, 14, 5], type=pyarrow.int64()),
    ]
    column_names = ["id", "name", "mass", "diameter", "density", "gravity", "escapeVelocity", "rotationPeriod", "lengthOfDay", "distanceFromSun", "perihelion", "aphelion", "orbitalPeriod", "orbitalVelocity", "orbitalInclination", "orbitalEccentricity", "obliquityToOrbit", "meanTemperature", "surfacePressure", "numberOfMoons"]

    # fmt: on
    full_set = pyarrow.Table.from_arrays(data, column_names)

    # make planet data act like it support temporality
    mask = [True, True, True, True, True, True, True, True, True]
    if end_date < datetime.date(1930, 3, 13):
        # March 13, 1930 - Pluto discovered by Clyde William Tombaugh
        mask = [True, True, True, True, True, True, True, True, False]
    if end_date < datetime.date(1846, 11, 13):
        # November 13, 1846 - Neptune
        mask = [True, True, True, True, True, True, True, False, False]
    if end_date < datetime.date(1781, 4, 26):
        # April 26, 1781 - Uranus discovered by Sir William Herschel
        mask = [True, True, True, True, True, True, False, False, False]

    return full_set.filter(mask)


schema = RelationSchema(
    table_name="$planets",
    columns=[
        FlatColumn(name="id", data_type=OPTERYX_TYPES.INTEGER),
        FlatColumn(name="name", data_type=OPTERYX_TYPES.VARCHAR),
        FlatColumn(name="mass", data_type=OPTERYX_TYPES.DOUBLE),
        FlatColumn(name="diameter", data_type=OPTERYX_TYPES.INTEGER),
        FlatColumn(name="density", data_type=OPTERYX_TYPES.INTEGER),
        FlatColumn(name="gravity", data_type=OPTERYX_TYPES.DOUBLE),
        FlatColumn(name="escapeVelocity", data_type=OPTERYX_TYPES.DOUBLE),
        FlatColumn(name="rotationPeriod", data_type=OPTERYX_TYPES.DOUBLE),
        FlatColumn(name="lengthOfDay", data_type=OPTERYX_TYPES.DOUBLE),
        FlatColumn(name="distanceFromSun", data_type=OPTERYX_TYPES.DOUBLE),
        FlatColumn(name="perihelion", data_type=OPTERYX_TYPES.DOUBLE),
        FlatColumn(name="aphelion", data_type=OPTERYX_TYPES.DOUBLE),
        FlatColumn(name="orbitalPeriod", data_type=OPTERYX_TYPES.DOUBLE),
        FlatColumn(name="orbitalVelocity", data_type=OPTERYX_TYPES.DOUBLE),
        FlatColumn(name="orbitalInclination", data_type=OPTERYX_TYPES.DOUBLE),
        FlatColumn(name="orbitalEccentricity", data_type=OPTERYX_TYPES.DOUBLE),
        FlatColumn(name="obliquityToOrbit", data_type=OPTERYX_TYPES.DOUBLE),
        FlatColumn(name="meanTemperature", data_type=OPTERYX_TYPES.DOUBLE),
        FlatColumn(name="surfacePressure", data_type=OPTERYX_TYPES.DOUBLE),
        FlatColumn(name="numberOfMoons", data_type=OPTERYX_TYPES.INTEGER),
    ],
)
